#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 20/02/2023, 14:04. Copyright (c) The Contributors

import pytest
from astropy.cosmology import Planck15
from astropy.units import Quantity

from xga.sourcetools.misc import nh_lookup, rad_to_ang, ang_to_rad, name_to_coord, coord_to_name
from .. import A907_LOC


@pytest.mark.heasoft
def test_nh_value():
    """
    Tests whether the nh_lookup tool is correctly interfacing with nh command and retrieving values. It possible
    that these values will change depending on the heasoft version (as different hydrogen maps are used), so a failure
    due to that reason isn't critical.
    """
    ret_val = nh_lookup(A907_LOC)
    assert ret_val[0].value == 0.0534 and ret_val[1].value == 0.0545


@pytest.mark.simple
def test_rad2ang():
    """
    Tests whether the conversion of a proper radius to angular radius is correct with a given redshift and cosmology.
    Not sure why I did this one as I can't imagine it ever being anything but true.
    """
    ret_val = rad_to_ang(Quantity(1000, 'kpc'), 0.3, Planck15)
    assert ret_val == Quantity(0.06046667460469272, 'deg')


@pytest.mark.simple
def test_ang2rad():
    """
    Tests whether the conversion of a angular radius to proper radius is correct with a given redshift and cosmology.
    Not sure why I did this one as I can't imagine it ever being anything but true.
    """
    ret_val = ang_to_rad(Quantity(0.06046667460469272, 'deg'), 0.3, Planck15)
    assert ret_val == Quantity(1000., 'kpc')


# I expect this to fail right now because I haven't finished it, I need to replace this once I have done it
@pytest.mark.xfail
@pytest.mark.simple
def test_name2coord():
    name_to_coord("XMMXCS J041853.9+555333.7")


@pytest.mark.simple
@pytest.mark.parametrize("test_coord, test_survey, expected", [(A907_LOC, 'XMMXCS', 'XMMXCSJ095822.1-110334.9'),
                                                               (A907_LOC, None, 'J095822.1-110334.9')])
def test_coord2name(test_coord, test_survey, expected):
    assert coord_to_name(test_coord, test_survey) == expected






#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 01/08/2024, 17:36. Copyright (c) The Contributors

import os
from random import randint
from shutil import rmtree
from typing import Union

import numpy as np
from astropy.units import Quantity, deg
from tqdm import tqdm

from .misc import cifbuild
from .run import sas_call
from .. import OUTPUT, NUM_CORES
from ..exceptions import SASInputInvalid, NoProductAvailableError
from ..imagetools import data_limits
from ..samples.base import BaseSample
from ..sources import BaseSource
from ..sources.base import NullSource
from ..utils import energy_to_channel


# TODO Perhaps remove the option to add to the SAS expression
@sas_call
def evselect_image(sources: Union[BaseSource, NullSource, BaseSample], lo_en: Quantity = Quantity(0.5, 'keV'),
                   hi_en: Quantity = Quantity(2.0, 'keV'), add_expr: str = "", num_cores: int = NUM_CORES,
                   disable_progress: bool = False):
    """
    A convenient Python wrapper for a configuration of the SAS evselect command that makes images.
    Images will be generated for every observation associated with every source passed to this function.
    If images in the requested energy band are already associated with the source,
    they will not be generated again.

    :param BaseSource/NullSource/BaseSample sources: A single source object, or a sample of sources.
    :param Quantity lo_en: The lower energy limit for the image, in astropy energy units.
    :param Quantity hi_en: The upper energy limit for the image, in astropy energy units.
    :param str add_expr: A string to be added to the SAS expression keyword
    :param int num_cores: The number of cores to use, default is set to 90% of available.
    :param bool disable_progress: Setting this to true will turn off the SAS generation progress bar.
    """
    stack = False  # This tells the sas_call routine that this command won't be part of a stack
    execute = True  # This should be executed immediately
    # This function supports passing both individual sources and sets of sources
    if isinstance(sources, (BaseSource, NullSource)):
        sources = [sources]

    # Don't do much value checking in this module, but this one is so fundamental that I will do it
    if lo_en > hi_en:
        raise ValueError("The 'lo_en' argument cannot be greater than 'hi_en'.")
    else:
        # Calls a useful little function that takes an astropy energy quantity to the XMM channels
        # required by SAS commands
        lo_chan = energy_to_channel(lo_en)
        hi_chan = energy_to_channel(hi_en)

    expr = " && ".join([e for e in ["expression='(PI in [{l}:{u}])".format(l=lo_chan, u=hi_chan),
                                    add_expr] if e != ""]) + "'"
    # These lists are to contain the lists of commands/paths/etc for each of the individual sources passed
    # to this function
    sources_cmds = []
    sources_paths = []
    sources_extras = []
    sources_types = []
    for source in sources:
        cmds = []
        final_paths = []
        extra_info = []
        # Check which event lists are associated with each individual source
        for pack in source.get_products("events", just_obj=False):
            obs_id = pack[0]
            inst = pack[1]

            if not os.path.exists(OUTPUT + obs_id):
                os.mkdir(OUTPUT + obs_id)

            en_id = "bound_{l}-{u}".format(l=lo_en.value, u=hi_en.value)
            exists = [match for match in source.get_products("image", obs_id, inst, just_obj=False)
                      if en_id in match]
            if len(exists) == 1 and exists[0][-1].usable:
                continue

            evt_list = pack[-1]
            dest_dir = OUTPUT + "{o}/{i}_{l}-{u}_{n}_temp/".format(o=obs_id, i=inst, l=lo_en.value, u=hi_en.value,
                                                                   n=source.name)
            im = "{o}_{i}_{l}-{u}keVimg.fits".format(o=obs_id, i=inst, l=lo_en.value, u=hi_en.value)

            # If something got interrupted and the temp directory still exists, this will remove it
            if os.path.exists(dest_dir):
                rmtree(dest_dir)

            os.makedirs(dest_dir)
            cmds.append("cd {d};evselect table={e} imageset={i} xcolumn=X ycolumn=Y ximagebinsize=87 "
                        "yimagebinsize=87 squarepixels=yes ximagesize=512 yimagesize=512 imagebinning=binSize "
                        "ximagemin=3649 ximagemax=48106 withxranges=yes yimagemin=3649 yimagemax=48106 "
                        "withyranges=yes {ex}; mv * ../; cd ..; rm -r {d}".format(d=dest_dir, e=evt_list.path,
                                                                                  i=im, ex=expr))

            # This is the products final resting place, if it exists at the end of this command
            final_paths.append(os.path.join(OUTPUT, obs_id, im))
            extra_info.append({"lo_en": lo_en, "hi_en": hi_en, "obs_id": obs_id, "instrument": inst})
        sources_cmds.append(np.array(cmds))
        sources_paths.append(np.array(final_paths))
        # This contains any other information that will be needed to instantiate the class
        # once the SAS cmd has run
        sources_extras.append(np.array(extra_info))
        sources_types.append(np.full(sources_cmds[-1].shape, fill_value="image"))

    # I only return num_cores here so it has a reason to be passed to this function, really
    # it could just be picked up in the decorator.
    return sources_cmds, stack, execute, num_cores, sources_types, sources_paths, sources_extras, disable_progress


@sas_call
def eexpmap(sources: Union[BaseSource, NullSource, BaseSample], lo_en: Quantity = Quantity(0.5, 'keV'),
            hi_en: Quantity = Quantity(2.0, 'keV'), num_cores: int = NUM_CORES, disable_progress: bool = False):
    """
    A convenient Python wrapper for the SAS eexpmap command.
    Expmaps will be generated for every observation associated with every source passed to this function.
    If expmaps in the requested energy band are already associated with the source,
    they will not be generated again.

    :param BaseSource/NullSource/BaseSample sources: A single source object, or sample of sources.
    :param Quantity lo_en: The lower energy limit for the expmap, in astropy energy units.
    :param Quantity hi_en: The upper energy limit for the expmap, in astropy energy units.
    :param int num_cores: The number of cores to use (if running locally), default is set to
        90% of available.
    :param bool disable_progress: Setting this to true will turn off the SAS generation progress bar.
    """
    # I know that a lot of this code is the same as the evselect_image code, but its 1am so please don't
    #  judge me too much.

    # This function supports passing both individual sources and sets of sources
    if isinstance(sources, (BaseSource, NullSource)):
        sources = [sources]

    # Don't do much value checking in this module, but this one is so fundamental that I will do it
    if lo_en > hi_en:
        raise ValueError("lo_en cannot be greater than hi_en")
    else:
        # Calls a useful little function that takes an astropy energy quantity to the XMM channels
        # required by SAS commands
        lo_chan = energy_to_channel(lo_en)
        hi_chan = energy_to_channel(hi_en)

    # These are crucial, to generate an exposure map one must have a ccf.cif calibration file, and a reference
    # image. If they do not already exist, these commands should generate them.
    cifbuild(sources, disable_progress=disable_progress, num_cores=num_cores)
    sources = evselect_image(sources, lo_en, hi_en)
    # This is necessary because the decorator will reduce a one element list of source objects to a single
    # source object. Useful for the user, not so much here where the code expects an iterable.
    if not isinstance(sources, (list, BaseSample)):
        sources = [sources]

    # These lists are to contain the lists of commands/paths/etc for each of the individual sources passed
    # to this function
    sources_cmds = []
    sources_paths = []
    sources_extras = []
    sources_types = []
    for source in sources:
        cmds = []
        final_paths = []
        extra_info = []
        # Check which event lists are associated with each individual source
        for pack in source.get_products("events", just_obj=False):
            obs_id = pack[0]
            inst = pack[1]

            if not os.path.exists(OUTPUT + obs_id):
                os.mkdir(OUTPUT + obs_id)

            en_id = "bound_{l}-{u}".format(l=lo_en.value, u=hi_en.value)
            exists = [match for match in source.get_products("expmap", obs_id, inst, just_obj=False)
                      if en_id in match]
            if len(exists) == 1 and exists[0][-1].usable:
                continue
            # Generating an exposure map requires a reference image.
            ref_im = [match for match in source.get_products("image", obs_id, inst, just_obj=False)
                      if en_id in match][0][-1]
            # It also requires an attitude file
            att = source.get_att_file(obs_id)
            # Set up the paths and names of files
            evt_list = pack[-1]
            dest_dir = OUTPUT + "{o}/{i}_{l}-{u}_{n}_temp/".format(o=obs_id, i=inst, l=lo_en.value, u=hi_en.value,
                                                                   n=source.name)
            exp_map = "{o}_{i}_{l}-{u}keVexpmap.fits".format(o=obs_id, i=inst, l=lo_en.value, u=hi_en.value)

            # If something got interrupted and the temp directory still exists, this will remove it
            if os.path.exists(dest_dir):
                rmtree(dest_dir)

            os.makedirs(dest_dir)
            cmds.append("cd {d}; cp ../ccf.cif .; export SAS_CCF={ccf}; eexpmap eventset={e} "
                        "imageset={im} expimageset={eim} withdetcoords=no withvignetting=yes "
                        "attitudeset={att} pimin={l} pimax={u}; mv * ../; cd ..; "
                        "rm -r {d}".format(e=evt_list.path, im=ref_im.path, eim=exp_map, att=att, l=lo_chan,
                                           u=hi_chan, d=dest_dir, ccf=dest_dir + "ccf.cif"))

            # This is the products final resting place, if it exists at the end of this command
            final_paths.append(os.path.join(OUTPUT, obs_id, exp_map))
            extra_info.append({"lo_en": lo_en, "hi_en": hi_en, "obs_id": obs_id, "instrument": inst})
        sources_cmds.append(np.array(cmds))
        sources_paths.append(np.array(final_paths))
        # This contains any other information that will be needed to instantiate the class
        # once the SAS cmd has run
        sources_extras.append(np.array(extra_info))
        sources_types.append(np.full(sources_cmds[-1].shape, fill_value="expmap"))

    stack = False  # This tells the sas_call routine that this command won't be part of a stack
    execute = True  # This should be executed immediately
    # I only return num_cores here so it has a reason to be passed to this function, really
    # it could just be picked up in the decorator.
    return sources_cmds, stack, execute, num_cores, sources_types, sources_paths, sources_extras, disable_progress


@sas_call
def emosaic(sources: Union[BaseSource, BaseSample], to_mosaic: str, lo_en: Quantity = Quantity(0.5, 'keV'),
            hi_en: Quantity = Quantity(2.0, 'keV'), psf_corr: bool = False, psf_model: str = "ELLBETA",
            psf_bins: int = 4, psf_algo: str = "rl", psf_iter: int = 15, num_cores: int = NUM_CORES,
            disable_progress: bool = False):
    """
    A convenient Python wrapper for the SAS emosaic command. Every image associated with the source,
    that is in the energy band specified by the user, will be added together.

    :param BaseSource/BaseSample sources: A single source object, or a sample of sources.
    :param str to_mosaic: The data type to produce a mosaic for, can be either image or expmap.
    :param Quantity lo_en: The lower energy limit for the combined image, in astropy energy units.
    :param Quantity hi_en: The upper energy limit for the combined image, in astropy energy units.
    :param bool psf_corr: If True, PSF corrected images will be mosaiced.
    :param str psf_model: If PSF corrected, the PSF model used.
    :param int psf_bins: If PSF corrected, the number of bins per side.
    :param str psf_algo: If PSF corrected, the algorithm used.
    :param int psf_iter: If PSF corrected, the number of algorithm iterations.
    :param int num_cores: The number of cores to use (if running locally), default is set to
        90% of available.
    :param bool disable_progress: Setting this to true will turn off the SAS generation progress bar.
    """
    # This function supports passing both individual sources and sets of sources
    if isinstance(sources, BaseSource):
        sources = [sources]

    # NullSources are not allowed to be mosaiced, as they can have any observations associated and thus won't
    #  necessarily overlap
    if isinstance(sources, NullSource):
        raise TypeError("You cannot create combined images of a NullSource")

    if to_mosaic not in ["image", "expmap"]:
        raise ValueError("The only valid choices for to_mosaic are image and expmap.")
    # Don't do much value checking in this module, but this one is so fundamental that I will do it
    elif lo_en > hi_en:
        raise ValueError("lo_en cannot be greater than hi_en")

    # To make a mosaic we need to have the individual products in the first place
    if to_mosaic == "image":
        sources = evselect_image(sources, lo_en, hi_en, disable_progress=disable_progress, num_cores=num_cores)
        for_name = "img"
    elif to_mosaic == "expmap":
        sources = eexpmap(sources, lo_en, hi_en, disable_progress=disable_progress, num_cores=num_cores)
        for_name = "expmap"

    # This is necessary because the decorator will reduce a one element list of source objects to a single
    # source object. Useful for the user, not so much here where the code expects an iterable.
    if not isinstance(sources, (list, BaseSample)):
        sources = [sources]

    # The bit on the end takes everything up out of the temporary folder and removes it
    mosaic_cmd = "cd {d}; emosaic imagesets='{ims}' mosaicedset={mim}; mv * ../; cd ..; rm -r {d}"

    sources_cmds = []
    sources_paths = []
    sources_extras = []
    sources_types = []
    for source in sources:
        en_id = "bound_{l}-{u}".format(l=lo_en.value, u=hi_en.value)
        # If we're mosaicing PSF corrected images, we need to
        if psf_corr and to_mosaic == "expmap":
            raise ValueError("There can be no PSF corrected expmaps to mosaic, it doesn't make sense.")
        elif psf_corr:
            en_id += "_" + psf_model + "_" + str(psf_bins) + "_" + psf_algo + str(psf_iter)

        # Checking if the combined product already exists
        exists = [match for match in source.get_products("combined_{}".format(to_mosaic), just_obj=False)
                  if en_id in match]
        if len(exists) == 1 and exists[0][-1].usable:
            sources_cmds.append(np.array([]))
            sources_paths.append(np.array([]))
            sources_extras.append(np.array([]))
            sources_types.append(np.array([]))
            continue

        # This fetches all image objects with the passed energy bounds
        matches = [[match[0], match[-1]] for match in source.get_products(to_mosaic, just_obj=False)
                   if en_id in match]

        # In theory this should never be triggered, because we already ran evselect_image so the images should be
        #  there - but I am now somehow having an error where we get to this point with no errors and no images, so
        #  we're going to add this in to be absolutely sure (as otherwise emosaic fails with a very unhelpful error).
        if len(matches) == 0:
            assoc = ", ".join([cur_oi + cur_i for cur_oi in source.instruments
                               for cur_i in source.instruments[cur_oi]])
            raise NoProductAvailableError("The images required for emosaic are not available for {p} - this is not a"
                                          " usual behaviour as XGA should have generated them; the relevant "
                                          "observations are {d}.".format(p=source.name, d=assoc))

        paths = [product[1].path for product in matches if product[1].usable]
        obs_ids = [product[0] for product in matches if product[1].usable]
        obs_ids_set = []
        for obs_id in obs_ids:
            if obs_id not in obs_ids_set:
                obs_ids_set.append(obs_id)

            if not os.path.exists(OUTPUT + obs_id):
                os.mkdir(OUTPUT + obs_id)

        # The files produced by this function will now be stored in the combined directory.
        final_dest_dir = OUTPUT + "combined/"
        rand_ident = randint(0, int(1e+8))
        # Makes absolutely sure that the random integer hasn't already been used
        while len([f for f in os.listdir(final_dest_dir) if str(rand_ident) in f.split(OUTPUT+"combined/")[-1]]) != 0:
            rand_ident = randint(0, int(1e+8))

        dest_dir = os.path.join(final_dest_dir, "temp_emosaic_{}".format(rand_ident))
        os.mkdir(dest_dir)

        # The name of the file used to contain all the ObsIDs that went into the stacked image/expmap. However
        #  this caused problems when too many ObsIDs were present and the filename was longer than allowed. So
        #  now I use the random identity I generated, and store the ObsID/instrument information in the inventory
        #  file
        if not psf_corr:
            mosaic = "{os}_{l}-{u}keVmerged_{t}.fits".format(os=rand_ident, l=lo_en.value, u=hi_en.value, t=for_name)
        else:
            mosaic = "{os}_{b}bin_{it}iter_{m}mod_{a}algo_{l}-{u}keVpsfcorr_merged_img." \
                     "fits".format(os=rand_ident, l=lo_en.value, u=hi_en.value, b=psf_bins, it=psf_iter, a=psf_algo,
                                   m=psf_model)

        sources_cmds.append(np.array([mosaic_cmd.format(ims=" ".join(paths), mim=mosaic, d=dest_dir)]))
        sources_paths.append(np.array([os.path.join(final_dest_dir, mosaic)]))
        # This contains any other information that will be needed to instantiate the class
        # once the SAS cmd has run
        # The 'combined' values for obs and inst here are crucial, they will tell the source object that the final
        # product is assigned to that these are merged products - combinations of all available data
        sources_extras.append(np.array([{"lo_en": lo_en, "hi_en": hi_en, "obs_id": "combined",
                                         "instrument": "combined", "psf_corr": psf_corr, "psf_algo": psf_algo,
                                         "psf_model": psf_model, "psf_iter": psf_iter, "psf_bins": psf_bins}]))
        sources_types.append(np.full(sources_cmds[-1].shape, fill_value=to_mosaic))

    stack = False  # This tells the sas_call routine that this command won't be part of a stack
    execute = True  # This should be executed immediately
    # I only return num_cores here so it has a reason to be passed to this function, really
    # it could just be picked up in the decorator.
    return sources_cmds, stack, execute, num_cores, sources_types, sources_paths, sources_extras, disable_progress


@sas_call
def psfgen(sources: Union[BaseSource, BaseSample], bins: int = 4, psf_model: str = "ELLBETA",
           num_cores: int = NUM_CORES, disable_progress: bool = False):
    """
    A wrapper for the psfgen SAS task. Used to generate XGA PSF objects, which in turn can be used to correct
    XGA images/ratemaps for optical effects. By default we use the ELLBETA model reported in Read et al. 2011
    (doi:10.1051/0004-6361/201117525), and generate a grid of binsxbins PSFs that can be used
    to correct for the PSF over an entire image. The energy dependence of the PSF is assumed to be minimal, and the
    resultant PSF object will be paired up with an image that matches it's ObsID and instrument.

    :param BaseSource/BaseSample sources: A single source object, or a sample of sources.
    :param int bins: The image coordinate space will be divided into a grid of size binsxbins, PSFs will be
        generated at the central coordinates of the grid chunks.
    :param str psf_model: Which model to use when generating the PSF, default is ELLBETA, the best available.
    :param int num_cores: The number of cores to use (if running locally), default is set to
        90% of available.
    :param bool disable_progress: Setting this to true will turn off the SAS generation progress bar.
    """
    stack = False  # This tells the sas_call routine that this command won't be part of a stack
    execute = True  # This should be executed immediately

    psf_model = psf_model.upper()
    allowed_models = ["ELLBETA", "LOW", "MEDIUM", "EXTENDED", "HIGH"]
    if psf_model not in allowed_models:
        raise SASInputInvalid("{0} is not a valid PSF model. Allowed models are "
                              "{1}".format(psf_model, ", ".join(allowed_models)))

    # Need a valid CIF for this task, so run cifbuild first
    cifbuild(sources, disable_progress=disable_progress, num_cores=num_cores)

    # This is necessary because the decorator will reduce a one element list of source objects to a single
    # source object. Useful for the user, not so much here where the code expects an iterable.
    if not isinstance(sources, (list, BaseSample)):
        sources = [sources]

    # NullSources are not allowed to be mosaiced, as they can have any observations associated and thus won't
    #  necessarily overlap
    if isinstance(sources, NullSource):
        raise NotImplementedError("You cannot currently use PSFGen with a NullSource.")

    with tqdm(desc='Preparing PSF generation commands', total=len(sources),
              disable=len(sources) == 0) as psfgen_prep_progress:
        # These lists are to contain the lists of commands/paths/etc for each of the individual sources passed
        # to this function
        sources_cmds = []
        sources_paths = []
        sources_extras = []
        sources_types = []
        for source in sources:
            cmds = []
            final_paths = []
            extra_info = []
            # Check which event lists are associated with each individual source
            for pack in source.get_products("events", just_obj=False):
                obs_id = pack[0]
                inst = pack[1]

                if not os.path.exists(OUTPUT + obs_id):
                    os.mkdir(OUTPUT + obs_id)

                # This looks for any image for this ObsID, instrument combo - it does assume that whatever
                #  it finds will be the same resolution as any images in other energy bands that XGA will
                #  create in the future.
                images = source.get_products("image", obs_id, inst, just_obj=True)

                if len(images) == 0:
                    raise NoProductAvailableError("There is no image available for {o} {i}, please generate "
                                                  "images before PSFs".format(o=obs_id, i=inst))

                # Checking if the Image products are the same shape that XGA makes
                res_match = [im for im in images if im.shape == (512, 512)]
                if len(res_match) == 0:
                    raise NoProductAvailableError("There is an image associated with {o} {i}, but it doesn't"
                                                  " appear to be at the resolution XGA uses - this is not "
                                                  "supported yet.")
                else:
                    image = res_match[0]

                # Here we try and find if this PSF configuration has already been run and has been
                #  associated with the source. If so then don't do it again.
                psfs = source.get_products("psf", obs_id, inst, extra_key=psf_model + "_" + str(bins))
                if len(psfs) != 0:
                    continue

                # This part is where we decide on the RA DEC coordinates for the centres of each
                #  PSF in our grid
                # This function gives us x and y limits for where there is data in an image, they are used as start
                #  and end coordinates for our bins so the PSFs are more focused on where there is actually data.
                x_lims, y_lims = data_limits(image)
                # Simple calculation to calculate step size in pixels, so how long each chunk will be in
                #  x and y directions
                x_step = (x_lims[1] - x_lims[0]) / bins
                y_step = (y_lims[1] - y_lims[0]) / bins

                # These are the x and y bin centre coordinates - when converted to RA and DEC this is where the
                #  PSF is generated at.
                x_cen_coords = np.arange(*x_lims, x_step) + (x_step / 2)
                y_cen_coords = np.arange(*y_lims, y_step) + (y_step / 2)

                # Get all combinations of the central coordinates using meshgrid, then turn them into
                #  an N row, 2 column numpy array of pixel coordinates for easy conversion to RA-DEC.
                pix_mesh = np.meshgrid(x_cen_coords, y_cen_coords)
                pix_coords = Quantity(np.stack([pix_mesh[0].ravel(), pix_mesh[1].ravel()]).T, 'pix')

                # But I also want to know the boundaries of the bins so I can easily select which parts of
                #  the image belong with each PSF in the grid
                x_boundaries = np.linspace(*x_lims, bins+1)
                y_boundaries = np.linspace(*y_lims, bins+1)

                # These two arrays give the x and y boundaries of the bins in the same order as the pix_coords array
                x_bound_coords = np.tile(np.stack([x_boundaries[0: -1].ravel(), x_boundaries[1:].ravel()]).T,
                                         (bins, 1))
                x_bound_coords = x_bound_coords.round(0).astype(int)

                y_bound_coords = np.repeat(np.stack([y_boundaries[0: -1].ravel(), y_boundaries[1:].ravel()]).T,
                                           bins, 0)
                y_bound_coords = y_bound_coords.round(0).astype(int)

                ra_dec_coords = image.coord_conv(pix_coords, deg)

                dest_dir = OUTPUT + "{o}/{i}_{n}_temp/".format(o=obs_id, i=inst, n=source.name)
                psf = "{o}_{i}_{b}bin_{m}mod_{ra}_{dec}_psf.fits"

                # The change directory and SAS setup commands
                init_cmd = "cd {d}; cp ../ccf.cif .; export SAS_CCF={ccf}; ".format(d=dest_dir,
                                                                                    ccf=dest_dir + "ccf.cif")

                # If something got interrupted and the temp directory still exists, this will remove it
                if os.path.exists(dest_dir):
                    rmtree(dest_dir)

                os.makedirs(dest_dir)

                psf_files = []
                total_cmd = init_cmd
                for pair_ind in range(ra_dec_coords.shape[0]):
                    # The coordinates at which this PSF will be generated
                    ra, dec = ra_dec_coords[pair_ind, :].value

                    psf_file = psf.format(o=obs_id, i=inst, b=bins, ra=ra, dec=dec, m=psf_model)
                    psf_files.append(os.path.join(OUTPUT, obs_id, psf_file))
                    # Going with xsize and ysize as 400 pixels, I think its enough and quite a bit faster than 1000
                    total_cmd += "psfgen image={i} coordtype=EQPOS level={m} energy=1000 xsize=400 ysize=400 x={ra} " \
                                 "y={dec} output={p}; ".format(i=image.path, m=psf_model, ra=ra, dec=dec, p=psf_file)

                total_cmd += "mv * ../; cd ..; rm -r {d}".format(d=dest_dir)
                cmds.append(total_cmd)
                # This is the products final resting place, if it exists at the end of this command
                # In this case it just checks for the final PSF in the grid, all other files in the grid
                # get stored in extra info.
                final_paths.append(os.path.join(OUTPUT, obs_id, psf_file))
                extra_info.append({"obs_id": obs_id, "instrument": inst, "model": psf_model, "chunks_per_side": bins,
                                   "files": psf_files, "x_bounds": x_bound_coords, "y_bounds": y_bound_coords})

            sources_cmds.append(np.array(cmds))
            sources_paths.append(np.array(final_paths))
            # This contains any other information that will be needed to instantiate the class
            # once the SAS cmd has run
            sources_extras.append(np.array(extra_info))
            sources_types.append(np.full(sources_cmds[-1].shape, fill_value="psf"))

            psfgen_prep_progress.update(1)

    # I only return num_cores here so it has a reason to be passed to this function, really
    # it could just be picked up in the decorator.
    return sources_cmds, stack, execute, num_cores, sources_types, sources_paths, sources_extras, disable_progress





#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 24/07/2024, 16:16. Copyright (c) The Contributors

import os
from random import randint
from typing import Union

import numpy as np
from fitsio import read_header

from .run import sas_call
from .. import OUTPUT, NUM_CORES
from ..exceptions import InvalidProductError
from ..samples.base import BaseSample
from ..sources import BaseSource
from ..sources.base import NullSource


@sas_call
def cifbuild(sources: Union[BaseSource, NullSource, BaseSample], num_cores: int = NUM_CORES,
             disable_progress: bool = False):
    """
    A wrapper for the XMM cifbuild command, which will be run before many of the more complex SAS commands, to
    check that a CIF compatible with the local version of SAS is available. The observation date is taken from an
    event list for a given ObsID, and the analysis date is set to the date which this function is run.

    :param BaseSource/NullSource/BaseSample sources: A single source object, or a sample of sources.
    :param int num_cores: The number of cores to use (if running locally), default is set to
        90% of available.
    :param bool disable_progress: Setting this to true will turn off the SAS generation progress bar.
    """
    # This function supports passing both individual sources and sets of sources
    if isinstance(sources, (BaseSource, NullSource)):
        sources = [sources]

    # This string contains the bash code to run cifbuild
    cif_cmd = "cd {d}; cifbuild calindexset=ccf.cif withobservationdate=yes " \
              "observationdate={od} ; mv * ../; cd ..; rm -r {n}"

    sources_cmds = []
    sources_paths = []
    sources_extras = []
    sources_types = []
    for source in sources:
        cmds = []
        final_paths = []
        extra_info = []
        for obs_id in source.obs_ids:
            # Fetch an events list for this ObsID, doesn't matter which
            some_evt_lists = source.get_products("events", obs_id=obs_id)
            obs_date = None
            for evt in some_evt_lists:
                # Reads in the header of the events list file
                evt_head = read_header(evt.path)
                # Then extracts the observation date, this is what we need to give cifbuild
                if "DATE-OBS" in evt_head:
                    obs_date = evt_head['DATE-OBS']
                    del evt_head
                    break
                else:
                    del evt_head

            if obs_date is None:
                raise InvalidProductError("All event lists for {} are missing the DATE-OBS header, this is required to"
                                          " run the cifbuild function.".format(obs_id))

            if not os.path.exists(OUTPUT + obs_id):
                os.mkdir(OUTPUT + obs_id)

            dest_dir = "{out}{obs}/".format(out=OUTPUT, obs=obs_id)
            temp_name = "tempdir_{}".format(randint(0, int(1e+8)))
            temp_dir = dest_dir + temp_name + "/"

            final_path = dest_dir + "ccf.cif"
            if not os.path.exists(final_path):
                if not os.path.exists(temp_dir):
                    os.makedirs(temp_dir)
                cmds.append(cif_cmd.format(d=temp_dir, od=obs_date, n=temp_name))
                final_paths.append(final_path)
                extra_info.append({})  # This doesn't need any extra information

        sources_cmds.append(np.array(cmds))
        sources_paths.append(np.array(final_paths))
        sources_extras.append(np.array(extra_info))
        sources_types.append(np.full(sources_cmds[-1].shape, fill_value="ccf"))

    stack = False  # This tells the sas_call routine that this command won't be part of a stack
    execute = True  # This should be executed immediately

    return sources_cmds, stack, execute, num_cores, sources_types, sources_paths, sources_extras, disable_progress






#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 02/08/2024, 14:32. Copyright (c) The Contributors


from typing import Tuple

import numpy as np
from astropy.cosmology import Cosmology
from astropy.units import Quantity, UnitBase, pix, deg, arcsec, UnitConversionError

from .misc import pix_deg_scale, pix_rad_to_physical, physical_rad_to_pix, rad_to_ang
from .. import DEFAULT_COSMO
from ..products import Image, RateMap
from ..products.profile import SurfaceBrightness1D


def annular_mask(centre: Quantity, inn_rad: np.ndarray, out_rad: np.ndarray, shape: tuple,
                 start_ang: Quantity = Quantity(0, 'deg'), stop_ang: Quantity = Quantity(360, 'deg')) -> np.ndarray:
    """
    A handy little function to generate annular (or circular) masks in the form of numpy arrays.
    It produces the src_mask for a given shape of image, centered at supplied coordinates, and with inner and
    outer radii supplied by the user also. Angular limits can also be supplied to give the src_mask an annular
    dependence. This function should be properly vectorised, and accepts inner and outer radii in
    the form of arrays. The result will be an len_y, len_x, N dimensional array, where N is equal to
    the length of inn_rad.

    :param Quantity centre: Astropy pix quantity of the form Quantity([x, y], pix).
    :param np.ndarray inn_rad: Pixel radius for the inner part of the annular src_mask.
    :param np.ndarray out_rad: Pixel radius for the outer part of the annular src_mask.
    :param Quantity start_ang: Lower angular limit for the src_mask.
    :param Quantity stop_ang: Upper angular limit for the src_mask.
    :param tuple shape: The output from the shape property of the numpy array you are generating masks for.
    :return: The generated src_mask array.
    :rtype: np.ndarray
    """
    # Split out the centre coordinates
    cen_x = centre[0].value
    cen_y = centre[1].value

    # Making use of the astropy units module, check that we are being pass actual angle values
    if start_ang.unit not in ['deg', 'rad']:
        raise ValueError("start_angle unit type {} is not an accepted angle unit, "
                         "please use deg or rad.".format(start_ang.unit))
    elif stop_ang.unit not in ['deg', 'rad']:
        raise ValueError("stop_angle unit type {} is not an accepted angle unit, "
                         "please use deg or rad.".format(stop_ang.unit))
    # Enforcing some common sense rules on the angles
    elif start_ang >= stop_ang:
        raise ValueError("start_ang cannot be greater than or equal to stop_ang.")
    elif start_ang > Quantity(360, 'deg') or stop_ang > Quantity(360, 'deg'):
        raise ValueError("start_ang and stop_ang cannot be greater than 360 degrees.")
    elif stop_ang < Quantity(0, 'deg'):
        raise ValueError("stop_ang cannot be less than 0 degrees.")
    else:
        # Don't want to pass astropy objects to numpy functions, but do need the angles in radians
        start_ang = start_ang.to('rad').value
        stop_ang = stop_ang.to('rad').value

    # Check that if the inner and outer radii are arrays, they're the same length
    if isinstance(inn_rad, (np.ndarray, list)) and len(inn_rad) != len(out_rad):
        raise ValueError("inn_rad and out_rad are not the same length")
    elif isinstance(inn_rad, list) and len(inn_rad) == len(out_rad):
        # If it is a list, just quickly transform to a numpy array
        inn_rad = np.array(inn_rad)
        out_rad = np.array(out_rad)

    # This sets up the cartesian coordinate grid of x and y values
    arr_y, arr_x = np.ogrid[:shape[0], :shape[1]]

    # Go to polar coordinates
    rec_x = arr_x - cen_x
    rec_y = arr_y - cen_y
    # Leave this as r**2 to avoid square rooting and involving floats
    init_r_squared = rec_x**2 + rec_y**2

    # arctan2 does just perform arctan on two values, but then uses the signs of those values to
    # decide the quadrant of the output
    init_arr_theta = (np.arctan2(rec_x, rec_y) - start_ang) % (2*np.pi)  # Normalising to 0-2pi range

    # If the radius limits are an array, the arrays that describe the space we have constructed are copied
    #  into a third dimension - This allows masks for different radii to be generated in a vectorised fashion
    if isinstance(inn_rad, np.ndarray):
        arr_r_squared = np.repeat(init_r_squared[:, :, np.newaxis], len(inn_rad), axis=2)
        arr_theta = np.repeat(init_arr_theta[:, :, np.newaxis], len(inn_rad), axis=2)
    else:
        arr_r_squared = init_r_squared
        arr_theta = init_arr_theta

    # This will deal properly with inn_rad and out_rads that are arrays
    if np.greater(inn_rad, out_rad).any():
        raise ValueError("inn_rad value cannot be greater than out_rad")
    else:
        rad_mask = (arr_r_squared < out_rad ** 2) & (arr_r_squared >= inn_rad ** 2)

    # Finally, puts a cut on the allowed angle, and combined the radius and angular cuts into the final src_mask
    ang_mask = arr_theta <= (stop_ang - start_ang)
    ann_mask = rad_mask * ang_mask

    # Should ensure that the central pixel will be 0 for annular masks that are bounded by zero.
    #  Sometimes they aren't because of custom angle choices
    if 0 in inn_rad:
        where_zeros = np.where(inn_rad == 0)[0]
        ann_mask[cen_y, cen_x, where_zeros] = 1

    if ann_mask.shape[-1] == 1:
        ann_mask = np.squeeze(ann_mask)

    # Returns the annular src_mask(s), in the form of a len_y, len_x, N dimension np array
    return ann_mask


def ann_radii(im_prod: Image, centre: Quantity, rad: Quantity, z: float = None, pix_step: int = 1,
              rad_units: UnitBase = arcsec, cosmo: Cosmology = DEFAULT_COSMO, min_central_pix_rad: int = 3,
              start_pix_rad: int = 0) -> Tuple[np.ndarray, np.ndarray, Quantity]:
    """
    Will probably only ever be called by an internal brightness calculation, but two different methods
    need it so it gets its own method.

    :param Image im_prod: An Image or RateMap product object that you wish to calculate annuli for.
    :param Quantity centre: The coordinates of the centre of the set of annuli.
    :param Quantity rad: The outer radius of the set of annuli.
    :param float z: The redshift of the source of interest, required if the output radius units are
        a proper radius.
    :param int pix_step: The width (in pixels) of each annular bin, default is 1.
    :param UnitBase rad_units: The output units for the centres of the annulli returned by
        this function. The inner and outer radii will always be in pixels.
    :param Cosmology cosmo: An instance of an astropy cosmology, the default is a concordance flat LambdaCDM model.
    :param int start_pix_rad: The pixel radius at which the innermost annulus starts, default is zero.
    :param int min_central_pix_rad: The minimum radius of the innermost circular annulus (will only
        be used if start_pix_rad is 0, otherwise the innermost annulus is not a circle), default is three.
    :return: Returns the inner and outer radii of the annuli (in pixels), and the centres of the annuli
        in cen_rad_units.
    :rtype: Tuple[np.ndarray, np.ndarray, Quantity]
    """
    # Quickly get the central coordinates in degrees as well
    deg_cen = im_prod.coord_conv(centre, deg)
    pix_cen = im_prod.coord_conv(centre, pix)

    # If the radius is passed in a proper distance unit, this will convert it or raise an exception
    #  if it can't
    if rad.unit.is_equivalent('deg') or rad.unit.is_equivalent('kpc'):
        rad = physical_rad_to_pix(im_prod, rad, deg_cen, z, cosmo)
    elif rad.unit == pix:
        # If the radius is already in pixels then all is well
        pass
    else:
        raise UnitConversionError('{} is not a recognised distance unit'.format(rad.unit.to_string()))

    rad = np.ceil(rad)

    # I'M LEAVING THIS HERE IN CASE I DECIDE TO RESTORE IT IN THE FUTURE - it was intended as a sort of
    #  safety feature, to ensure that the central was big enough that it wasn't going to be zero valued. Unfortunately
    #  it really screws over inverse abel transforming the data, because many of the techniques implemented in PyAbel
    #  require uniform radial (in my case) sampling - that can't be true if this feature is in unless the pix_step
    #  is set to at least 3 - enforcing that as a minimum may well be useful, but for now I'm disabling this

    # if pix_step < 3 and start_pix_rad == 0:
    #     central_circ = np.array([0, min_central_pix_rad], dtype=int)
    #     ann_rads = np.arange(min_central_pix_rad+pix_step, rad.value + 1, pix_step).astype(int)
    #     rads = np.concatenate([central_circ, ann_rads])

    # By this point, the rad should be in pixels
    rads = np.arange(start_pix_rad, rad.value + 1, pix_step).astype(int)

    inn_rads = rads[:len(rads) - 1]
    out_rads = rads[1:len(rads)]

    pix_cen_rads = Quantity((inn_rads + out_rads)/2, pix)
    cen_rads = pix_rad_to_physical(im_prod, pix_cen_rads, rad_units, deg_cen, z, cosmo)

    # If the innermost radius is zero then the innermost annulus is actually a circle and we don't want
    #  the central radius to be between it and the next radius, as that wouldn't be strictly accurate
    if rads[0] == 0:
        cen_rads[0] = Quantity(0, rad_units)

    return inn_rads, out_rads, cen_rads


def radial_brightness(rt: RateMap, centre: Quantity, outer_rad: Quantity, back_inn_rad_factor: float = 1.05,
                      back_out_rad_factor: float = 1.5, interloper_mask: np.ndarray = None,
                      z: float = None, pix_step: int = 1, rad_units: UnitBase = arcsec,
                      cosmo: Cosmology = DEFAULT_COSMO, min_snr: float = 0.0, min_central_pix_rad: int = 3,
                      start_pix_rad: int = 0) -> Tuple[SurfaceBrightness1D, bool]:
    """
    A simple method to calculate the average brightness in circular annuli upto the radius of
    the chosen region. The annuli are one pixel in width, and as this uses the masks that were generated
    earlier, interloper sources should be removed.

    :param RateMap rt: A RateMap object to construct a brightness profile from.
    :param Quantity centre: The coordinates for the centre of the brightness profile.
    :param Quantity outer_rad: The outer radius of the brightness profile.
    :param float back_inn_rad_factor: This factor is multiplied by the outer pixel radius, which gives the inner
        radius for the background mask.
    :param float back_out_rad_factor: This factor is multiplied by the outer pixel radius, which gives the outer
        radius for the background mask.
    :param np.ndarray interloper_mask: A numpy array that masks out any interloper sources.
    :param float z: The redshift of the source of interest.
    :param int pix_step: The width (in pixels) of each annular bin, default is 1.
    :param BaseUnit rad_units: The desired output units for the central radii of the annuli.
    :param Cosmology cosmo: An astropy cosmology object for source coordinate conversions.
    :param float min_snr: The minimum signal to noise allowed for each bin in the profile. If any point is
        below this threshold the profile will be rebinned. Default is 0.0
    :param int start_pix_rad: The pixel radius at which the innermost annulus starts, default is zero.
    :param int min_central_pix_rad: The minimum radius of the innermost circular annulus (will only
        be used if start_pix_rad is 0, otherwise the innermost annulus is not a circle), default is three.
    :return: The brightness is returned in a flat numpy array, then the radii at the centre of the bins are
        returned in units of kpc, the width of the bins, and finally the average brightness in the background region is
        returned.
    :rtype: Tuple[SurfaceBrightness1D, bool]
    """

    def _iterative_profile(annulus_masks: np.ndarray, inner_rads: np.ndarray, outer_rads: np.ndarray) \
            -> Tuple[bool, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Internal function to calculate and re-bin (ONCE) a surface brightness profile. The profile, along with
        modified (if rebinned) masks and radii arrays are returned to the user. This can be called once, or iteratively
        by a loop.

        :param np.ndarray annulus_masks: 512x512xN numpy array of annular masks, where N is the number of annuli
        :param np.ndarray inner_rads: The inner radii of the annuli.
        :param np.ndarray outer_rads: The outer radii of the annuli.
        :return: Boolean variable that describes whether another re-binning iteration is required, the
            brightness profile and uncertainties, the modified annular masks, inner radii, outer radii, and annulus areas.
        :rtype:
        """
        # These are annular masks with interloper sources removed, sensor and edge masks applied
        corr_ann_masks = (annulus_masks * interloper_mask[..., None] * rt.sensor_mask[..., None]
                          * rt.edge_mask[..., None])

        # This calculates the area of each annulus mask
        num_pix = np.sum(corr_ann_masks, axis=(0, 1))
        ann_areas = num_pix * to_arcmin ** 2

        # Applying the annular masks, with interlopers removed, and chip edges
        masked_countrate_data = corr_ann_masks * rt.data[..., None]
        masked_countrate_error_data = corr_ann_masks * count_rate_err_map[..., None]

        # Defining the brightness profile with the current annuli
        bright_profile = np.sum(masked_countrate_data, axis=(0, 1)) / ann_areas

        # And the uncertainties on the profiles. Adding pixels in quadrature
        bright_profile_errors = np.sqrt(np.sum(masked_countrate_error_data**2, axis=(0, 1))) / ann_areas

        # We want to check if all parts of the profile are above the defined minimum signal to noise
        snr_prof = bright_profile / countrate_bg_per_area

        # Checking if we are below the minimum signal to noise anywhere
        below = np.argwhere(snr_prof < min_snr).flatten()
        if below.shape != (0,) and below[0] != (bright_profile.shape[0] - 1):
            annulus_masks[:, :, below[0]] = annulus_masks[:, :, below[0]] + annulus_masks[:, :, below[0] + 1]
            annulus_masks = np.delete(annulus_masks, below[0] + 1, axis=2)

            outer_rads[below[0]] = outer_rads[below[0] + 1]
            outer_rads = np.delete(outer_rads, below[0] + 1)
            inner_rads = np.delete(inner_rads, below[0] + 1)

            another_pass = True

        elif below.shape != (0,) and below[0] == (bright_profile.shape[0] - 1):
            annulus_masks[:, :, below[0] - 1] = annulus_masks[:, :, below[0] - 1] + annulus_masks[:, :, below[0]]
            annulus_masks = np.delete(annulus_masks, below[0], axis=2)

            outer_rads[below[0] - 1] = outer_rads[below[0]]
            outer_rads = np.delete(outer_rads, below[0])
            inner_rads = np.delete(inner_rads, below[0])

            another_pass = True

        else:
            another_pass = False

        return another_pass, bright_profile, bright_profile_errors, annulus_masks, inner_rads, outer_rads, ann_areas

    if interloper_mask is not None and rt.shape != interloper_mask.shape:
        raise ValueError("The shape of the src_mask array {0} must be the same as that of im_prod "
                         "{1}.".format(interloper_mask.shape, rt.shape))
    elif interloper_mask is None:
        interloper_mask = np.ones(rt.shape)

    # Returns conversion factor to degrees, so multiplying by 60 goes to arcminutes
    # Getting this because we want to be able to convert pixel distance into arcminutes for dividing by the area
    to_arcmin = pix_deg_scale(centre, rt.radec_wcs).value * 60

    # Just making sure we have the centre in pixel coordinates
    pix_cen = rt.coord_conv(centre, pix)

    # This sets up the initial annular bin radii, as well as finding the central radii of the bins in the chosen units.
    inn_rads, out_rads, cen_rads = ann_radii(rt, centre, outer_rad, z, pix_step, rad_units, cosmo,
                                             min_central_pix_rad, start_pix_rad)

    # These calculate the inner and out pixel radii for the background mask - placed in arrays because the
    #  annular mask function expects iterable radii. As pixel radii have to be integer for generating a mask,
    #  I've used ceil.
    back_inn_rad = np.array([np.ceil(out_rads[-1]*back_inn_rad_factor).astype(int)])
    back_out_rad = np.array([np.ceil(out_rads[-1]*back_out_rad_factor).astype(int)])

    # Using my annular mask function to make a nice background region, which will be corrected for instrumental
    #  stuff and interlopers in a second
    back_mask = annular_mask(pix_cen, back_inn_rad, back_out_rad, rt.shape)

    # Includes chip gaps, interloper removal, and edge removal (to try and avoid artificially bright pixels)
    #  in the background mask
    corr_back_mask = back_mask * rt.sensor_mask * rt.edge_mask * interloper_mask

    # Calculates the area of the background region in arcmin^2
    back_pix = np.sum(corr_back_mask, axis=(0, 1))
    back_area = back_pix * to_arcmin**2
    if back_area == 0:
        raise ValueError("The background mask combined with the sensor mask is is all zeros, this is probably"
                         " because you're looking at a large cluster at a low redshift.")
    # Per pix is the background which has been divided by the number of pixels
    count_bg_per_pix = np.sum(rt.image.data * corr_back_mask, axis=(0, 1)) / back_pix
    # This is where the pixels have been converted into an actual area measurement in arcmin units
    countrate_bg_per_area = np.sum(rt.data * corr_back_mask, axis=(0, 1)) / back_area

    # Defined here so we can use the where argument in np.sqrt
    err_calc = (rt.image.data - count_bg_per_pix) + 2*count_bg_per_pix
    # Errors on the raw photon counts
    count_err_map = np.zeros(rt.shape)
    # Turning those into count rate errors
    np.sqrt(err_calc, where=err_calc > 0, out=count_err_map)

    # I divide by a copy of the expmap data array here, otherwise it breaks, and if I'm honest I don't know why...
    count_rate_err_map = np.divide(count_err_map, rt.expmap.data.copy(), where=rt.expmap.data != 0) * rt.edge_mask

    # Using the ellipse adds enough : to get all the dimensions in the array, then the None adds an empty
    #  dimension. Helps with broadcasting the annular masks with the region src_mask that gets rid of interlopers
    ann_masks = annular_mask(pix_cen, inn_rads, out_rads, rt.shape)

    # Make copies of originals in case re-binning fails
    init_inn = inn_rads.copy()
    init_out = out_rads.copy()

    # This gets switched to false if the signal to noise conditions are not fulfilled
    calculate_profile = True
    while calculate_profile:
        calculate_profile, br_prof, br_errs, ann_masks, inn_rads, \
            out_rads, areas = _iterative_profile(ann_masks, inn_rads, out_rads)

    if len(inn_rads) == 0:
        # This means that rebinning failed, and the profile couldn't be made greater than the minimum signal to noise
        inn_rads = init_inn
        out_rads = init_out
        ann_masks = annular_mask(pix_cen, inn_rads, out_rads, rt.shape)
        # Set the min_snr to 0 to get an non-rebinned profile
        min_snr = 0
        prof_results = _iterative_profile(ann_masks, inn_rads, out_rads)
        br_prof, br_errs = prof_results[1:3]
        areas = prof_results[-1]
        succeeded = False
    else:
        succeeded = True

    final_inn_rads = pix_rad_to_physical(rt, Quantity(inn_rads, pix), rad_units, centre, z, cosmo)
    final_out_rads = pix_rad_to_physical(rt, Quantity(out_rads, pix), rad_units, centre, z, cosmo)

    # Need these simply because the brightness profile must always be aware of the radii in degrees in order to
    #  to assemble a storage key properly
    deg_inn_rads = pix_rad_to_physical(rt, Quantity(inn_rads, pix), deg, centre, z, cosmo)
    deg_out_rads = pix_rad_to_physical(rt, Quantity(out_rads, pix), deg, centre, z, cosmo)

    cen_rads = (final_inn_rads + final_out_rads) / 2
    deg_cen_rads = (deg_inn_rads + deg_out_rads) / 2
    rad_err = (final_out_rads-final_inn_rads) / 2

    if not outer_rad.unit.is_equivalent('deg'):
        deg_outer_rad = rad_to_ang(outer_rad, z, cosmo).to('deg')
    else:
        deg_outer_rad = outer_rad.to("deg")

    # Now I've finally implemented some profile product classes I can just smoosh everything into a convenient product
    br_prof = SurfaceBrightness1D(rt, cen_rads, Quantity(br_prof, 'ct/(s*arcmin**2)'), centre, pix_step, min_snr,
                                  deg_outer_rad, rad_err, Quantity(br_errs, 'ct/(s*arcmin**2)'),
                                  Quantity(countrate_bg_per_area, 'ct/(s*arcmin**2)'),
                                  np.insert(out_rads, 0, inn_rads[0]), np.concatenate([back_inn_rad, back_out_rad]),
                                  Quantity(areas, 'arcmin**2'), deg_cen_rads, succeeded, True)

    return br_prof, succeeded


# TODO REWRITE THESE AT SOME POINT
def pizza_brightness(im_prod: Image, src_mask: np.ndarray, back_mask: np.ndarray,
                     centre: Quantity, rad: Quantity, num_slices: int = 4,
                     z: float = None, pix_step: int = 1, cen_rad_units: UnitBase = arcsec,
                     cosmo=DEFAULT_COSMO) -> Tuple[np.ndarray, Quantity, Quantity, np.float64, np.ndarray, np.ndarray]:

    raise NotImplementedError("The supporting infrastructure to allow pizza profile product objects hasn't been"
                              " written yet sorry!")





#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 08/06/2023, 22:40. Copyright (c) The Contributors

from typing import Union, List

import numpy as np
from astropy.units import Quantity, Unit, UnitConversionError, kpc, deg

from .base import BaseModel1D
from ..utils import r500, r200, r2500


class KingProfile1D(BaseModel1D):
    """
    An XGA model implementation of the King profile, describing an isothermal sphere. This describes a
    radial density profile and assumes spherical symmetry.

    :param Unit/str x_unit: The unit of the x-axis of this model, kpc for instance. May be passed as a string
        representation or an astropy unit object.
    :param Unit/str y_unit: The unit of the output of this model, keV for instance. May be passed as a string
        representation or an astropy unit object.
    :param List[Quantity] cust_start_pars: The start values of the model parameters for any fitting function that
        used start values. The units are checked against default start values.
    """
    def __init__(self, x_unit: Union[str, Unit] = 'kpc', y_unit: Union[str, Unit] = Unit('Msun/Mpc^3'),
                 cust_start_pars: List[Quantity] = None):
        """
        The init of a subclass of the XGA BaseModel1D class, describing a basic model for galaxy cluster gas
        density, the king profile.
        """
        # If a string representation of a unit was passed then we make it an astropy unit
        if isinstance(x_unit, str):
            x_unit = Unit(x_unit)
        if isinstance(y_unit, str):
            y_unit = Unit(y_unit)

        poss_y_units = [Unit('Msun/Mpc^3'), Unit('1/cm^3')]
        y_convertible = [u.is_equivalent(y_unit) for u in poss_y_units]
        if not any(y_convertible):
            allowed = ", ".join([u.to_string() for u in poss_y_units])
            raise UnitConversionError("{p} is not convertible to any of the allowed units; "
                                      "{a}".format(p=y_unit.to_string(), a=allowed))
        else:
            yu_ind = y_convertible.index(True)

        poss_x_units = [kpc, deg, r200, r500, r2500]
        x_convertible = [u.is_equivalent(x_unit) for u in poss_x_units]
        if not any(x_convertible):
            allowed = ", ".join([u.to_string() for u in poss_x_units])
            raise UnitConversionError("{p} is not convertible to any of the allowed units; "
                                      "{a}".format(p=x_unit.to_string(), a=allowed))
        else:
            xu_ind = x_convertible.index(True)

        r_core_starts = [Quantity(100, 'kpc'), Quantity(0.2, 'deg'), Quantity(0.05, r200), Quantity(0.1, r500),
                         Quantity(0.5, r2500)]
        # TODO MAKE THE NEW START PARAMETERS MORE SENSIBLE
        norm_starts = [Quantity(1e+13, 'Msun/Mpc^3'), Quantity(1e-3, '1/cm^3')]
        start_pars = [Quantity(1, ''), r_core_starts[xu_ind], norm_starts[yu_ind]]
        if cust_start_pars is not None:
            # If the custom start parameters can run this gauntlet without tripping an error then we're all good
            # This method also returns the custom start pars converted to exactly the same units as the default
            start_pars = self.compare_units(cust_start_pars, start_pars)

        r_core_priors = [{'prior': Quantity([0, 2000], 'kpc'), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], 'deg'), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r200), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r500), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r2500), 'type': 'uniform'}]
        norm_priors = [{'prior': Quantity([1e+12, 1e+16], 'Msun/Mpc^3'), 'type': 'uniform'},
                       {'prior': Quantity([0, 10], '1/cm^3'), 'type': 'uniform'}]

        priors = [{'prior': Quantity([0, 3]), 'type': 'uniform'}, r_core_priors[xu_ind], norm_priors[yu_ind]]

        nice_pars = [r"$\beta$", r"R$_{\rm{core}}$", "N$_{0}$"]
        info_dict = {'author': 'placeholder', 'year': 'placeholder', 'reference': 'placeholder',
                     'general': 'The un-projected version of the beta profile, suitable for a simple fit\n'
                                ' to 3D density distributions. Describes a simple isothermal sphere.'}
        super().__init__(x_unit, y_unit, start_pars, priors, 'king', 'King Profile', nice_pars, 'Gas Density',
                         info_dict)

    @staticmethod
    def model(x: Quantity, beta: Quantity, r_core: Quantity, norm: Quantity) -> Quantity:
        """
        The model function for the king profile.

        :param Quantity x: The radii to calculate y values for.
        :param Quantity beta: The beta slope parameter of the model.
        :param Quantity r_core: The core radius.
        :param Quantity norm: The normalisation of the model.
        :return: The y values corresponding to the input x values.
        :rtype: Quantity
        """
        return norm * ((1 + (x / r_core)**2)**(-3 * beta))

    def derivative(self, x: Quantity, dx: Quantity = Quantity(0, ''), use_par_dist: bool = False) -> Quantity:
        """
        Calculates the gradient of the king profile at a given point, overriding the numerical method implemented
        in the BaseModel1D class, as this simple model has an easily derivable first derivative.

        :param Quantity x: The point(s) at which the slope of the model should be measured. If multiple,
            non-distribution, radii are to be used, make sure to pass them as an (M,), single dimension, astropy
            quantity, where M is the number of separate radii to generate realisations for. To marginalise over a
            radius distribution when generating realisations, pass a multi-dimensional astropy quantity; i.e. for
            a single set of realisations pass a (1, N) quantity, where N is the number of samples in the parameter
            posteriors, for realisations for M different radii pass a (M, N) quantity.
        :param Quantity dx: This makes no difference here, as this is an analytical derivative. It has
            been left in so that the inputs for this method don't vary between models.
        :param bool use_par_dist: Should the parameter distributions be used to calculate a derivative
            distribution; this can only be used if a fit has been performed using the model instance.
            Default is False, in which case the current parameters will be used to calculate a single value.
        :return: The calculated slope of the model at the supplied x position(s).
        :rtype: Quantity
        """
        # This makes sure that the input radius, or radii, are being used properly. For a single x value, or a set of
        #  x values, we use [..., None] to ensure that (if the user has decided to use the parameter distribution) we
        #  get M distributions of the derivative. In the case where the user has passed a distribution of radii, really
        #  representative of a single radius but with uncertainty, then this does not trigger, and rather than an N
        #  by N (where N is the number of samples in each posterior) array, you get an M x N array
        if x.isscalar or (not x.isscalar and x.ndim == 1):
            x = x[..., None]
        if not use_par_dist:
            beta, r_core, norm = self._model_pars
        else:
            beta, r_core, norm = self.par_dists

        return (-6*beta*norm*x/np.power(r_core, 2))*np.power((1+np.power(x/r_core, 2)), (-3*beta) - 1)


class DoubleKingProfile1D(BaseModel1D):
    """
    An XGA model implementation of the double King profile, simply the sum of two King profiles. This describes a
    radial density profile and assumes spherical symmetry.

    :param Unit/str x_unit: The unit of the x-axis of this model, kpc for instance. May be passed as a string
        representation or an astropy unit object.
    :param Unit/str y_unit: The unit of the output of this model, keV for instance. May be passed as a string
        representation or an astropy unit object.
    :param List[Quantity] cust_start_pars: The start values of the model parameters for any fitting function that
        used start values. The units are checked against default start values.
    """
    def __init__(self, x_unit: Union[str, Unit] = 'kpc', y_unit: Union[str, Unit] = Unit('Msun/Mpc^3'),
                 cust_start_pars: List[Quantity] = None):
        """
        The init of a subclass of the XGA BaseModel1D class, describing a basic model for galaxy cluster gas
        density, the king profile.
        """
        # If a string representation of a unit was passed then we make it an astropy unit
        if isinstance(x_unit, str):
            x_unit = Unit(x_unit)
        if isinstance(y_unit, str):
            y_unit = Unit(y_unit)

        poss_y_units = [Unit('Msun/Mpc^3'), Unit('1/cm^3')]
        y_convertible = [u.is_equivalent(y_unit) for u in poss_y_units]
        if not any(y_convertible):
            allowed = ", ".join([u.to_string() for u in poss_y_units])
            raise UnitConversionError("{p} is not convertible to any of the allowed units; "
                                      "{a}".format(p=y_unit.to_string(), a=allowed))
        else:
            yu_ind = y_convertible.index(True)

        poss_x_units = [kpc, deg, r200, r500, r2500]
        x_convertible = [u.is_equivalent(x_unit) for u in poss_x_units]
        if not any(x_convertible):
            allowed = ", ".join([u.to_string() for u in poss_x_units])
            raise UnitConversionError("{p} is not convertible to any of the allowed units; "
                                      "{a}".format(p=x_unit.to_string(), a=allowed))
        else:
            xu_ind = x_convertible.index(True)

        r_core_starts = [Quantity(100, 'kpc'), Quantity(0.2, 'deg'), Quantity(0.05, r200), Quantity(0.1, r500),
                         Quantity(0.5, r2500)]
        # TODO MAKE THE NEW START PARAMETERS MORE SENSIBLE
        norm_starts = [Quantity(1e+13, 'Msun/Mpc^3'), Quantity(1e-3, '1/cm^3')]
        start_pars = [Quantity(1, ''), r_core_starts[xu_ind], norm_starts[yu_ind],
                      Quantity(1, ''), r_core_starts[xu_ind], norm_starts[yu_ind]]
        if cust_start_pars is not None:
            # If the custom start parameters can run this gauntlet without tripping an error then we're all good
            # This method also returns the custom start pars converted to exactly the same units as the default
            start_pars = self.compare_units(cust_start_pars, start_pars)

        # TODO MAYBE ADJUST ALL OF THE PRIORS ETC AS THEY'RE JUST COPIED FROM KING
        r_core_priors = [{'prior': Quantity([0, 2000], 'kpc'), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], 'deg'), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r200), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r500), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r2500), 'type': 'uniform'}]
        norm_priors = [{'prior': Quantity([1e+12, 1e+16], 'Msun/Mpc^3'), 'type': 'uniform'},
                       {'prior': Quantity([0, 10], '1/cm^3'), 'type': 'uniform'}]

        priors = [{'prior': Quantity([0, 3]), 'type': 'uniform'}, r_core_priors[xu_ind], norm_priors[yu_ind],
                  {'prior': Quantity([0, 3]), 'type': 'uniform'}, r_core_priors[xu_ind], norm_priors[yu_ind]]

        nice_pars = [r"$\beta_{1}$", r"R$_{\rm{core}, 1}$", "N$_{0, 1}$", r"$\beta_{2}$", r"R$_{\rm{core}, 2}$",
                     "N$_{0, 2}$"]
        info_dict = {'author': 'placeholder', 'year': 'placeholder', 'reference': 'placeholder',
                     'general': 'placeholder'}
        super().__init__(x_unit, y_unit, start_pars, priors, 'double_king', 'Double King Profile', nice_pars,
                         'Gas Density', info_dict)

    @staticmethod
    def model(x: Quantity, beta_one: Quantity, r_core_one: Quantity, norm_one: Quantity, beta_two: Quantity,
              r_core_two: Quantity, norm_two: Quantity) -> Quantity:
        """
        The model function for the double King profile.

        :param Quantity x: The radii to calculate y values for.
        :param Quantity beta_one: The beta slope parameter of the first King model.
        :param Quantity r_core_one: The core radius of the first King model.
        :param Quantity norm_one: The normalisation of the first King model.
        :param Quantity beta_two: The beta slope parameter of the second King model.
        :param Quantity r_core_two: The core radius of the second King model.
        :param Quantity norm_two: The normalisation of the second King model.
        :return: The y values corresponding to the input x values.
        :rtype: Quantity
        """
        return (norm_one * ((1 + (x / r_core_one)**2)**(-3 * beta_one))) + \
               (norm_two * ((1 + (x / r_core_two)**2)**(-3 * beta_two)))

    def derivative(self, x: Quantity, dx: Quantity = Quantity(0, ''), use_par_dist: bool = False) -> Quantity:
        """
        Calculates the gradient of the double King profile at a given point, overriding the numerical method implemented
        in the BaseModel1D class, as this simple model has an easily derivable first derivative.

        :param Quantity x: The point(s) at which the slope of the model should be measured. If multiple,
            non-distribution, radii are to be used, make sure to pass them as an (M,), single dimension, astropy
            quantity, where M is the number of separate radii to generate realisations for. To marginalise over a
            radius distribution when generating realisations, pass a multi-dimensional astropy quantity; i.e. for
            a single set of realisations pass a (1, N) quantity, where N is the number of samples in the parameter
            posteriors, for realisations for M different radii pass a (M, N) quantity.
        :param Quantity dx: This makes no difference here, as this is an analytical derivative. It has
            been left in so that the inputs for this method don't vary between models.
        :param bool use_par_dist: Should the parameter distributions be used to calculate a derivative
            distribution; this can only be used if a fit has been performed using the model instance.
            Default is False, in which case the current parameters will be used to calculate a single value.
        :return: The calculated slope of the model at the supplied x position(s).
        :rtype: Quantity
        """
        # This makes sure that the input radius, or radii, are being used properly. For a single x value, or a set of
        #  x values, we use [..., None] to ensure that (if the user has decided to use the parameter distribution) we
        #  get M distributions of the derivative. In the case where the user has passed a distribution of radii, really
        #  representative of a single radius but with uncertainty, then this does not trigger, and rather than an N
        #  by N (where N is the number of samples in each posterior) array, you get an M x N array
        if x.isscalar or (not x.isscalar and x.ndim == 1):
            x = x[..., None]

        if not use_par_dist:
            beta_one, r_core_one, norm_one, beta_two, r_core_two, norm_two = self._model_pars
        else:
            beta_one, r_core_one, norm_one, beta_two, r_core_two, norm_two = self.par_dists
        p1 = (-6*beta_one*norm_one*x/np.power(r_core_one, 2))*np.power((1+np.power(x/r_core_one, 2)), (-3*beta_one) - 1)
        p2 = (-6*beta_two*norm_two*x/np.power(r_core_two, 2))*np.power((1+np.power(x/r_core_two, 2)), (-3*beta_two) - 1)
        return p1 + p2


class SimpleVikhlininDensity1D(BaseModel1D):
    """
    An XGA model implementation of a simplified version of Vikhlinin's full density model. Used relatively recently
    in https://doi.org/10.1051/0004-6361/201833325 by Ghirardini et al., a simplified form of Vikhlinin's full
    density model, which can be found in https://doi.org/10.1086/500288.

    :param Unit/str x_unit: The unit of the x-axis of this model, kpc for instance. May be passed as a string
        representation or an astropy unit object.
    :param Unit/str y_unit: The unit of the output of this model, keV for instance. May be passed as a string
        representation or an astropy unit object.
    :param List[Quantity] cust_start_pars: The start values of the model parameters for any fitting function that
        used start values. The units are checked against default start values.
    """
    def __init__(self, x_unit: Union[str, Unit] = 'kpc', y_unit: Union[str, Unit] = Unit('Msun/Mpc^3'),
                 cust_start_pars: List[Quantity] = None):
        """
        The init of a subclass of the XGA BaseModel1D class, describing a simplified version of Vikhlinin et al.'s
        model for the gas density profile of a galaxy cluster.
        """
        # If a string representation of a unit was passed then we make it an astropy unit
        if isinstance(x_unit, str):
            x_unit = Unit(x_unit)
        if isinstance(y_unit, str):
            y_unit = Unit(y_unit)

        poss_y_units = [Unit('Msun/Mpc^3'), Unit('1/cm^3')]
        y_convertible = [u.is_equivalent(y_unit) for u in poss_y_units]
        if not any(y_convertible):
            allowed = ", ".join([u.to_string() for u in poss_y_units])
            raise UnitConversionError("{p} is not convertible to any of the allowed units; "
                                      "{a}".format(p=y_unit.to_string(), a=allowed))
        else:
            yu_ind = y_convertible.index(True)

        poss_x_units = [kpc, deg, r200, r500, r2500]
        x_convertible = [u.is_equivalent(x_unit) for u in poss_x_units]
        if not any(x_convertible):
            allowed = ", ".join([u.to_string() for u in poss_x_units])
            raise UnitConversionError("{p} is not convertible to any of the allowed units; "
                                      "{a}".format(p=x_unit.to_string(), a=allowed))
        else:
            xu_ind = x_convertible.index(True)

        r_core_starts = [Quantity(100, 'kpc'), Quantity(0.2, 'deg'), Quantity(0.05, r200), Quantity(0.1, r500),
                         Quantity(0.5, r2500)]
        r_s_starts = [Quantity(300, 'kpc'), Quantity(0.35, 'deg'), Quantity(0.15, r200), Quantity(0.3, r500),
                      Quantity(1.0, r2500)]

        norm_starts = [Quantity(1e+13, 'Msun/Mpc^3'), Quantity(1e-3, '1/cm^3')]
        start_pars = [Quantity(1, ''), r_core_starts[xu_ind], Quantity(1, ''), r_s_starts[xu_ind], Quantity(2, ''),
                      norm_starts[yu_ind]]
        if cust_start_pars is not None:
            # If the custom start parameters can run this gauntlet without tripping an error then we're all good
            # This method also returns the custom start pars converted to exactly the same units as the default
            start_pars = self.compare_units(cust_start_pars, start_pars)

        r_core_priors = [{'prior': Quantity([0, 2000], 'kpc'), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], 'deg'), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r200), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r500), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r2500), 'type': 'uniform'}]
        norm_priors = [{'prior': Quantity([1e+12, 1e+16], 'Msun/Mpc^3'), 'type': 'uniform'},
                       {'prior': Quantity([0, 10], '1/cm^3'), 'type': 'uniform'}]

        priors = [{'prior': Quantity([0, 3]), 'type': 'uniform'}, r_core_priors[xu_ind],
                  {'prior': Quantity([0, 3]), 'type': 'uniform'}, r_core_priors[xu_ind],
                  {'prior': Quantity([0, 5]), 'type': 'uniform'}, norm_priors[yu_ind]]

        nice_pars = [r"$\beta$", r"R$_{\rm{core}}$", r"$\alpha$", r"R$_{\rm{s}}$", r"$\epsilon$", r"N$_{0}$"]
        info_dict = {'author': 'Ghirardini et al.', 'year': 2019,
                     'reference': 'https://doi.org/10.1051/0004-6361/201833325',
                     'general': "A simplified form of Vikhlinin's full density model, a type of broken\n"
                                " power law that deals well with most galaxy cluster density profile."}
        super().__init__(x_unit, y_unit, start_pars, priors, 'simple_vikhlinin_dens', 'Simplified Vikhlinin Profile',
                         nice_pars, 'Gas Density', info_dict)

    @staticmethod
    def model(x: Quantity, beta: Quantity, r_core: Quantity, alpha: Quantity, r_s: Quantity, epsilon: Quantity,
              norm: Quantity) -> Quantity:
        """
        The model function for the simplified Vikhlinin density profile.

        :param Quantity x: The radii to calculate y values for.
        :param Quantity beta: The beta parameter of the model.
        :param Quantity r_core: The core radius of the model.
        :param Quantity alpha: The alpha parameter of the model.
        :param Quantity r_s: The radius near where a change of slope by epsilon occurs.
        :param Quantity epsilon: The epsilon parameter of the model.
        :param Quantity norm: The overall normalisation of the model.
        :return: The y values corresponding to the input x values.
        :rtype: Quantity
        """
        try:
            # Calculates the ratio of the r_values to the r_core parameter
            rc_rat = x / r_core
            # Calculates the ratio of the r_values to the r_s parameter
            rs_rat = x / r_s

            first_term = rc_rat**(-alpha) / ((1+rc_rat**2)**((3 * beta) - (alpha / 2)))
            second_term = 1 / ((1 + rs_rat**3)**(epsilon / 3))
            result = norm * np.sqrt(first_term * second_term)
        except ZeroDivisionError:
            result = np.NaN

        return result

    def derivative(self, x: Quantity, dx: Quantity = Quantity(0, ''), use_par_dist: bool = False) -> Quantity:
        """
        Calculates the gradient of the simple Vikhlinin density profile at a given point, overriding the
        numerical method implemented in the BaseModel1D class.

        :param Quantity x: The point(s) at which the slope of the model should be measured. If multiple,
            non-distribution, radii are to be used, make sure to pass them as an (M,), single dimension, astropy
            quantity, where M is the number of separate radii to generate realisations for. To marginalise over a
            radius distribution when generating realisations, pass a multi-dimensional astropy quantity; i.e. for
            a single set of realisations pass a (1, N) quantity, where N is the number of samples in the parameter
            posteriors, for realisations for M different radii pass a (M, N) quantity.
        :param Quantity dx: This makes no difference here, as this is an analytical derivative. It has
            been left in so that the inputs for this method don't vary between models.
        :param bool use_par_dist: Should the parameter distributions be used to calculate a derivative
            distribution; this can only be used if a fit has been performed using the model instance.
            Default is False, in which case the current parameters will be used to calculate a single value.
        :return: The calculated slope of the model at the supplied x position(s).
        :rtype: Quantity
        """
        # This makes sure that the input radius, or radii, are being used properly. For a single x value, or a set of
        #  x values, we use [..., None] to ensure that (if the user has decided to use the parameter distribution) we
        #  get M distributions of the derivative. In the case where the user has passed a distribution of radii, really
        #  representative of a single radius but with uncertainty, then this does not trigger, and rather than an N
        #  by N (where N is the number of samples in each posterior) array, you get an M x N array
        if x.isscalar or (not x.isscalar and x.ndim == 1):
            x = x[..., None]

        if not use_par_dist:
            beta, r_core, alpha, r_s, epsilon, norm = self.model_pars
        else:
            beta, r_core, alpha, r_s, epsilon, norm = self.par_dists

        # TODO DOUBLE CHECK THIS WHEN I'M LESS TIRED
        first_term = -1*norm*np.sqrt(np.power(x/r_core, -alpha)*np.power((x**3/r_s**3) + 1, -epsilon/3)
                                     *np.power((x**2/r_core**2) + 1, 0.5*(alpha-(6*beta))))
        second_term = 1/(2*x*(x**2 + r_core**2)*(x**3 + r_s**3))
        third_term = (x**3 + r_s**3)*(6*beta*x**2 + alpha*r_core**2) + x**3*epsilon*(x**2 + r_core**2)

        return first_term*second_term*third_term


class VikhlininDensity1D(BaseModel1D):
    """
    An XGA model implementation of Vikhlinin's full density model for galaxy cluster intra-cluster medium,
    which can be found in https://doi.org/10.1086/500288. It is a radial profile, so an assumption
    of spherical symmetry is baked in.

    :param Unit/str x_unit: The unit of the x-axis of this model, kpc for instance. May be passed as a string
        representation or an astropy unit object.
    :param Unit/str y_unit: The unit of the output of this model, keV for instance. May be passed as a string
        representation or an astropy unit object.
    :param List[Quantity] cust_start_pars: The start values of the model parameters for any fitting function that
        used start values. The units are checked against default start values.
    """
    def __init__(self, x_unit: Union[str, Unit] = 'kpc', y_unit: Union[str, Unit] = Unit('Msun/Mpc^3'),
                 cust_start_pars: List[Quantity] = None):
        """
        The init of a subclass of the XGA BaseModel1D class, describing the full version of Vikhlinin et al.'s
        model for the gas density profile of a galaxy cluster.
        """

        # If a string representation of a unit was passed then we make it an astropy unit
        if isinstance(x_unit, str):
            x_unit = Unit(x_unit)
        if isinstance(y_unit, str):
            y_unit = Unit(y_unit)

        poss_y_units = [Unit('Msun/Mpc^3'), Unit('1/cm^3')]
        y_convertible = [u.is_equivalent(y_unit) for u in poss_y_units]
        if not any(y_convertible):
            allowed = ", ".join([u.to_string() for u in poss_y_units])
            raise UnitConversionError("{p} is not convertible to any of the allowed units; "
                                      "{a}".format(p=y_unit.to_string(), a=allowed))
        else:
            yu_ind = y_convertible.index(True)

        poss_x_units = [kpc, deg, r200, r500, r2500]
        x_convertible = [u.is_equivalent(x_unit) for u in poss_x_units]
        if not any(x_convertible):
            allowed = ", ".join([u.to_string() for u in poss_x_units])
            raise UnitConversionError("{p} is not convertible to any of the allowed units; "
                                      "{a}".format(p=x_unit.to_string(), a=allowed))
        else:
            xu_ind = x_convertible.index(True)

        r_core_starts = [Quantity(100, 'kpc'), Quantity(0.2, 'deg'), Quantity(0.05, r200), Quantity(0.1, r500),
                         Quantity(0.5, r2500)]
        r_s_starts = [Quantity(300, 'kpc'), Quantity(0.35, 'deg'), Quantity(0.15, r200), Quantity(0.3, r500),
                      Quantity(1.0, r2500)]
        r_core_two_starts = [Quantity(50, 'kpc'), Quantity(0.1, 'deg'), Quantity(0.03, r200), Quantity(0.05, r500),
                             Quantity(0.25, r2500)]

        norm_starts = [Quantity(1e+13, 'Msun/Mpc^3'), Quantity(1e-3, '1/cm^3')]
        norm_two_starts = [Quantity(5e+12, 'Msun/Mpc^3'), Quantity(5e-4, '1/cm^3')]
        start_pars = [Quantity(1, ''), r_core_starts[xu_ind], Quantity(1, ''), r_s_starts[xu_ind], Quantity(2, ''),
                      Quantity(3, ''), norm_starts[yu_ind], Quantity(1, ''), r_core_two_starts[xu_ind],
                      norm_two_starts[yu_ind]]
        if cust_start_pars is not None:
            # If the custom start parameters can run this gauntlet without tripping an error then we're all good
            # This method also returns the custom start pars converted to exactly the same units as the default
            start_pars = self.compare_units(cust_start_pars, start_pars)

        r_core_priors = [{'prior': Quantity([0, 2000], 'kpc'), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], 'deg'), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r200), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r500), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r2500), 'type': 'uniform'}]
        norm_priors = [{'prior': Quantity([[1e+12, 1e+16]], 'Msun/Mpc^3'), 'type': 'uniform'},
                       {'prior': Quantity([0, 10], '1/cm^3'), 'type': 'uniform'}]

        priors = [{'prior': Quantity([0, 3]), 'type': 'uniform'}, r_core_priors[xu_ind],
                  {'prior': Quantity([0, 3]), 'type': 'uniform'}, r_core_priors[xu_ind],
                  {'prior': Quantity([0, 5]), 'type': 'uniform'}, {'prior': Quantity([-5, 5]), 'type': 'uniform'},
                  norm_priors[yu_ind], {'prior': Quantity([0, 3]), 'type': 'uniform'}, r_core_priors[xu_ind],
                  norm_priors[yu_ind]]

        nice_pars = [r"$\beta_{1}$", r"R$_{\rm{core,1}}$", r"$\alpha$", r"R$_{\rm{s}}$", r"$\epsilon$", r"$\gamma$",
                     r"N$_{01}$", r"$\beta_{2}$", r"R$_{\rm{core,2}}$", r"N$_{02}$"]
        info_dict = {'author': 'Vikhlinin et al.', 'year': 2006,
                     'reference': 'https://doi.org/10.1086/500288',
                     'general': "The full model for cluster density profiles created by Vikhlinin et al.\n"
                                "This model has MANY free parameters which can be very hard to get constraints\n"
                                " on, and as such many people would use the simplified version which is implemented\n"
                                " as the SimpleVikhlininDensity1D class in XGA."}
        super().__init__(x_unit, y_unit, start_pars, priors, 'vikhlinin_dens', 'Vikhlinin Profile',
                         nice_pars, 'Gas Density', info_dict)

    @staticmethod
    def model(x: Quantity, beta_one: Quantity, r_core_one: Quantity, alpha: Quantity, r_s: Quantity, epsilon: Quantity,
              gamma: Quantity, norm_one: Quantity, beta_two: Quantity, r_core_two: Quantity, norm_two: Quantity):
        """
        The model function for the full Vikhlinin density profile.

        :param Quantity x: The radii to calculate y values for.
        :param Quantity beta_one: The beta parameter of the model.
        :param Quantity r_core_one: The core radius of the model.
        :param Quantity alpha: The alpha parameter of the model.
        :param Quantity r_s: The radius near where a change of slope by epsilon occurs.
        :param Quantity epsilon: The epsilon parameter of the model.
        :param Quantity gamma: Width of slope change transition region.
        :param Quantity norm_one: The normalisation of the model first part of the model.
        :param Quantity beta_two: The beta parameter slope of the small core part of the model.
        :param Quantity r_core_two:The core radius of the small core part of the model.
        :param Quantity norm_two: The normalisation of the additive, small core part of the model.
        """
        try:
            # Calculates the ratio of the r_values to the r_core_one parameter
            rc1_rat = x / r_core_one
            # Calculates the ratio of the r_values to the r_core_two parameter
            rc2_rat = x / r_core_two
            # Calculates the ratio of the r_values to the r_s parameter
            rs_rat = x / r_s

            first_term = rc1_rat**(-alpha) / ((1 + rc1_rat**2)**((3 * beta_one) - (alpha / 2)))
            second_term = 1 / ((1 + rs_rat**gamma)**(epsilon / gamma))
            additive_term = 1 / ((1 + rc2_rat**2)**(3 * beta_two))
        except ZeroDivisionError:
            return np.NaN

        return np.sqrt((np.power(norm_one, 2) * first_term * second_term) + (np.power(norm_two, 2) * additive_term))

    def derivative(self, x: Quantity, dx: Quantity = Quantity(0, ''), use_par_dist: bool = False) -> Quantity:
        """
        Calculates the gradient of the full Vikhlinin density profile at a given point, overriding the
        numerical method implemented in the BaseModel1D class.

        :param Quantity x: The point(s) at which the slope of the model should be measured. If multiple,
            non-distribution, radii are to be used, make sure to pass them as an (M,), single dimension, astropy
            quantity, where M is the number of separate radii to generate realisations for. To marginalise over a
            radius distribution when generating realisations, pass a multi-dimensional astropy quantity; i.e. for
            a single set of realisations pass a (1, N) quantity, where N is the number of samples in the parameter
            posteriors, for realisations for M different radii pass a (M, N) quantity.
        :param Quantity dx: This makes no difference here, as this is an analytical derivative. It has
            been left in so that the inputs for this method don't vary between models.
        :param bool use_par_dist: Should the parameter distributions be used to calculate a derivative
            distribution; this can only be used if a fit has been performed using the model instance.
            Default is False, in which case the current parameters will be used to calculate a single value.
        :return: The calculated slope of the model at the supplied x position(s).
        :rtype: Quantity
        """
        # This makes sure that the input radius, or radii, are being used properly. For a single x value, or a set of
        #  x values, we use [..., None] to ensure that (if the user has decided to use the parameter distribution) we
        #  get M distributions of the derivative. In the case where the user has passed a distribution of radii, really
        #  representative of a single radius but with uncertainty, then this does not trigger, and rather than an N
        #  by N (where N is the number of samples in each posterior) array, you get an M x N array
        if x.isscalar or (not x.isscalar and x.ndim == 1):
            x = x[..., None]
        if not use_par_dist:
            b, rc, a, rs, e, g, n, b2, rc2, n2 = self.model_pars
        else:
            b, rc, a, rs, e, g, n, b2, rc2, n2 = self.par_dists

        # Its horrible I know...
        p1 = (-6*b2*(n2**2)*x*(((x/rc2)**2) + 1)**((-3*b2)-1)) / rc2**2
        p2 = (-a*(n**2)*((x/rc)**(-a-1))*((((x/rc)**2)+1)**((a/2)-3*b))*((((x/rs)**g) + 1)**(-e/g)))/rc
        p3 = (2*(n**2)*x*((a/2)-(3*b))*((x/rc)**(-a))*((((x/rc)**2)+1)**((a/2)-(3*b)-1))*((((x/rs)**g) + 1)**(-e/g)))/rc**2
        p4 = -(n**2)*e*(x**(g-1))*(rs**(-g))*((x/rc)**(-a))*((((x/rc)**2)+1)**((a/2)-(3*b)))*((((x/rs)**g)+1)**(-e/g-1))
        p5 = 2*np.sqrt((n2**2)*((((x/rc2)**2)+1)**(-3*b2)) + (n**2)*((x/rc)**(-a))*((((x/rc)**2)+1)**((a/2)-(3*b)))*((((x/rs)**g)+1)**(-e/g)))

        return (p1 + p2 + p3 + p4) / p5


DENS_MODELS = {"simple_vikhlinin_dens": SimpleVikhlininDensity1D, 'king': KingProfile1D,
               'double_king': DoubleKingProfile1D, 'vikhlinin_dens': VikhlininDensity1D}
DENS_MODELS_PAR_NAMES = {n: m().par_publication_names for n, m in DENS_MODELS.items()}
DENS_MODELS_PUB_NAMES = {n: m().publication_name for n, m in DENS_MODELS.items()}



